#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup


setup(name='roam',
      version='0.3.1',
      description='Easily traverse nested Python data structures ',
      author='James Murty',
      author_email='james@murty.co',
      url='https://github.com/jmurty/roam/',
      py_modules=['roam'],
      python_requires='>=3.6',
     )
